# Licensed under the Apache License: http://www.apache.org/licenses/LICENSE-2.0
# For details: https://github.com/gaogaotiantian/viztracer/blob/master/NOTICE.txt

from .util import compare_version, color_print
from . import __version__


class VizPluginError(Exception):
    pass


# A third party developer who wants to develop based on VizTracer can do a plugin
# Simply inherit VizPluginBase class and finish the methods. Then you can load it
# by VizTracer(plugins=[YourVizPlugin()])

class VizPluginBase:
    def __init__(self):
        pass

    def support_version(self):
        # You have to overload this to return the latest version of viztracer
        # your plugin supports. This is for API backward compatibility.
        # Simply return the version string
        # For example:
        #     return "0.10.5"
        raise NotImplementedError("Plugin of viztracer has to implement support_version method")

    def message(self, m_type: str, payload: dict):
        """
        This is the only logical interface with VizTracer. To make it simple and flexible,
        we use m_type for message type, and the payload could be any json compatible
        data. This is more extensible in the future
        :param m_type str: the message type VizPlugin is receiving
        :param payload dict: payload of the message

        :return dict: always return a dict. Return None if nothing needs to be done
                      by VizTracer. Otherwise refer to the docs
        """
        if m_type == "command":
            if payload["cmd_type"] == "terminate":
                return {"success": True}

        return {}


class VizPluginManager:
    def __init__(self, tracer, plugins):
        self._tracer = tracer
        self._plugins = []
        for plugin in plugins:
            if isinstance(plugin, VizPluginBase):
                plugin_instance = plugin
            elif isinstance(plugin, str):
                plugin_instance = self._get_plugin_from_string(plugin)
            else:
                raise TypeError("Invalid plugin!")
            self._plugins.append(plugin_instance)
            self._send_message(plugin_instance, "event", {"when": "initialize"})

    def _get_plugin_from_string(self, plugin):
        args = plugin.split()
        module = args[0]
        try:
            package = __import__(module)
        except (ImportError, ModuleNotFoundError):
            print(f"There's no module named {module}, maybe you need to install it")
            exit(1)

        m = package
        if "." in module:
            # package.module
            names = module.split(".")

            try:
                for mod in names[1:]:
                    m = m.__getattribute__(mod)
            except AttributeError:  # pragma: no cover
                # This in theory should never happen
                raise ImportError(f"Unable to import {module}, wrong path")
        try:
            m = m.__getattribute__("get_vizplugin")
        except AttributeError:
            print(f"Unable to find get_vizplugin in {module}. Incorrect plugin.")
            exit(1)

        if callable(m):
            return m(plugin)
        else:
            print(f"Unable to find get_vizplugin as a callable in {module}. Incorrect plugin.")
            exit(1)

    def _send_message(self, plugin, m_type, payload):
        # this is the only interface to communicate with vizplugin
        # in the future we may need to do version compatibility
        # here
        support_version = plugin.support_version()
        if compare_version(support_version, __version__) > 0:
            color_print("WARNING", "The plugin support version is higher than "
                                   "viztracer version. Consider update your viztracer")

        ret = plugin.message(m_type, payload)
        if m_type == "command":
            self.assert_success(plugin, payload, ret)
        else:
            self.resolve(support_version, ret)

    def event(self, when):
        for plugin in self._plugins:
            self._send_message(plugin, "event", {"when": when})

    def command(self, cmd):
        for plugin in self._plugins:
            self._send_message(plugin, "command", cmd)

    def terminate(self):
        self.command({"cmd_type": "terminate"})
        for plugin in self._plugins:
            del plugin
        self._plugins = []

    def assert_success(self, plugin, cmd, ret):
        if not ret or "success" not in ret or not ret["success"]:
            raise VizPluginError(f"{plugin} failed to process {cmd}")

    def resolve(self, version, ret):
        if not ret or "action" not in ret:
            return
        if ret["action"] == "handle_data":
            ret["handler"](self._tracer.data)



# Licensed under the Apache License: http://www.apache.org/licenses/LICENSE-2.0
# For details: https://github.com/gaogaotiantian/viztracer/blob/master/NOTICE.txt

from logging import Handler


class VizLoggingHandler(Handler):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._tracer = None

    def emit(self, record):
        if not self._tracer:
            return
        self._tracer.add_instant("logging", {"data": self.format(record)}, scope="p")

    def setTracer(self, tracer):
        self._tracer = tracer



# Licensed under the Apache License: http://www.apache.org/licenses/LICENSE-2.0
# For details: https://github.com/gaogaotiantian/viztracer/blob/master/NOTICE.txt

import platform
import subprocess


def size_fmt(num, suffix='B'):
    for unit in ['', 'Ki', 'Mi', 'Gi']:
        if abs(num) < 1024.0:
            return "%3.1f%s%s" % (num, unit, suffix)
        num /= 1024.0
    return "%.1f%s%s" % (num, 'Ti', suffix)


class _bcolors:
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'


bcolors = _bcolors()


def color_print(color, s):
    print(bcolors.__getattribute__(color) + s + bcolors.ENDC)


def get_url_from_file(file_path):
    # file_path should be absolute
    uname = platform.uname()
    if uname.system.lower() == "linux" and "microsoft" in uname.release.lower():
        # We are on WSL, need to convert file_path
        p = subprocess.run(["wslpath", "-w", file_path], stdout=subprocess.PIPE)
        if p.returncode == 0:
            file_path = p.stdout.decode("utf-8")
        else:
            raise Exception("Can't convert path '{}' to Windows path".format(file_path))

    return file_path


def compare_version(ver1, ver2):
    # assuming ver1, ver2 are both str and in a pattern of
    # major.minor.micro with only numbers
    # return 1 if ver1 > ver2
    # return 0 if ver1 == ver2
    # return -1 if ver1 < ver2
    tuple1 = tuple((int(v) for v in ver1.split(".")))
    tuple2 = tuple((int(v) for v in ver2.split(".")))

    if tuple1 > tuple2:
        return 1
    elif tuple1 == tuple2:
        return 0
    else:
        return -1


def get_tracer():
    try:
        return __viz_tracer__
    except NameError:
        return None



# Licensed under the Apache License: http://www.apache.org/licenses/LICENSE-2.0
# For details: https://github.com/gaogaotiantian/viztracer/blob/master/NOTICE.txt

import os
try:
    import orjson as json
except ImportError:
    import json
from string import Template


class _FlameNode:
    def __init__(self, parent, name):
        self.name = name
        self.value = 0
        self.last_entry = -1
        self.last_exit = -1
        self.parent = parent
        self.children = {}

    def json(self):
        return {
            "name": self.name,
            "value": self.value,
            "children": [child.json() for child in self.children.values()]
        }


class _FlameTree:
    def __init__(self):
        self.root = _FlameNode(None, "__root__")
        self.curr = self.root

    def add_entry(self, data):
        if self.root.last_entry == -1:
            self.root.last_entry = data["ts"]
        if data["name"] in self.curr.children:
            self.curr = self.curr.children[data["name"]]
        else:
            node = _FlameNode(self.curr, data["name"])
            self.curr.children[data["name"]] = node
            self.curr = node
        self.curr.last_entry = data["ts"]

    def add_exit(self, data):
        if self.curr != self.root:
            self.curr.value += data["ts"] - self.curr.last_entry
            self.curr = self.curr.parent
            if self.curr == self.root:
                self.root.value = data["ts"] - self.root.last_entry

    def add_complete(self, data):
        if self.root.last_entry == -1:
            self.root.last_entry = data["ts"]

        # Find this nodes parent
        while self.curr.last_exit != -1 and self.curr.last_exit < data["ts"]:
            self.curr = self.curr.parent
        if data["name"] in self.curr.children:
            self.curr = self.curr.children[data["name"]]
        else:
            node = _FlameNode(self.curr, data["name"])
            self.curr.children[data["name"]] = node
            self.curr = node
        self.curr.value += data["dur"]
        self.curr.last_exit = data["ts"] + data["dur"]
        if self.curr.parent == self.root:
            self.root.value = data["ts"] + data["dur"] - self.root.last_entry

    def json(self):
        return self.root.json()


class FlameGraph:
    def __init__(self, trace_data=None):
        if trace_data:
            self._data = self.parse(trace_data)

    def parse(self, trace_data):
        trees = {}
        ret = {}
        for data in trace_data["traceEvents"]:
            key = "p{}_t{}".format(data["pid"], data["tid"])
            if key in trees:
                tree = trees[key]
            else:
                tree = _FlameTree()
                trees[key] = tree
            if data["ph"] == "B":
                tree.add_entry(data)
            elif data["ph"] == "E":
                tree.add_exit(data)
            elif data["ph"] == "X":
                tree.add_complete(data)
        for key in trees:
            ret[key] = trees[key].json()
        return ret

    def load(self, input_file):
        with open(input_file, encoding="utf-8") as f:
            self._data = self.parse(json.loads(f.read()))

    def save(self, output_file="result_flamegraph.html"):
        sub = {}
        with open(os.path.join(os.path.dirname(__file__), "html/flamegraph.html"), encoding="utf-8") as f:
            tmpl = f.read()
        sub["data"] = self._data

        with open(output_file, "w", encoding="utf-8") as f:
            f.write(Template(tmpl).substitute(sub))



# Licensed under the Apache License: http://www.apache.org/licenses/LICENSE-2.0
# For details: https://github.com/gaogaotiantian/viztracer/blob/master/NOTICE.txt

from string import Template
import os
import io
import re
import gzip
try:
    import orjson as json
except ImportError:
    import json
from .util import size_fmt, color_print


def get_json(data):
    # This function will return a json object if data is already json object
    # or a opened file or a file path
    if type(data) is dict:
        # This is an object already
        return data
    elif isinstance(data, io.IOBase):
        json_str = data.read()
    elif type(data) is str:
        with open(data, encoding="utf-8") as f:
            json_str = f.read()
    else:
        raise TypeError("Unexpected Type{}!", type(data))

    try:
        return json.loads(json_str)
    except Exception as e:
        print("Unable to decode {}".format(data))
        raise e


class ReportBuilder:
    def __init__(self, data, verbose=1):
        self.verbose = verbose
        self.combined_json = None
        self.entry_number_threshold = 4000000
        if type(data) is list:
            self.jsons = [get_json(j) for j in data]
        else:
            self.jsons = [get_json(data)]

    def combine_json(self):
        if self.combined_json:
            return 0
        if not self.jsons:
            raise Exception("Can't get report of nothing")
        self.combined_json = self.jsons[0]
        if len(self.jsons) > 1:
            for one in self.jsons[1:]:
                if "traceEvents" in one:
                    self.combined_json["traceEvents"].extend(one["traceEvents"])

    def generate_json(self, allow_binary=False, file_info=False):
        self.combine_json()
        if self.verbose > 0:
            entries = len(self.combined_json["traceEvents"])
            print(f"Dumping trace data to json, total entries: {entries}, estimated json file size: {size_fmt(120 * entries)}")
            if entries >= self.entry_number_threshold:
                print("")
                color_print("WARNING", "Large trace requires a lot of RAM and is slow to load.")
                color_print("WARNING", ("    If you need faster loading time or smaller trace file, "
                                        "try a smaller tracer_entries or use filters"))
                color_print("WARNING", "    use --quiet to shut me up")
                print("")

        if file_info:
            self.combined_json["file_info"] = {"files": {}, "functions": {}}
            pattern = re.compile(r".*\((.*):([0-9]*)\)")
            file_dict = self.combined_json["file_info"]["files"]
            func_dict = self.combined_json["file_info"]["functions"]
            for event in self.combined_json["traceEvents"]:
                if event["ph"] == 'X':
                    if event["name"] not in func_dict:
                        try:
                            m = pattern.match(event["name"])
                            file_name = m.group(1)
                            lineno = int(m.group(2))
                            if file_name not in file_dict:
                                with open(file_name, "r", encoding="utf-8") as f:
                                    content = f.read()
                                    file_dict[file_name] = [content, content.count("\n")]
                            func_dict[event["name"]] = [file_name, lineno]
                        except Exception:
                            pass

        if json.__name__ == "orjson":
            if allow_binary:
                return json.dumps(self.combined_json)
            else:
                return json.dumps(self.combined_json).decode("utf-8")
        else:
            return json.dumps(self.combined_json)

    def generate_report(self, file_info=False):
        sub = {}
        with open(os.path.join(os.path.dirname(__file__), "html/trace_viewer_embedder.html"), encoding="utf-8") as f:
            tmpl = f.read()
        with open(os.path.join(os.path.dirname(__file__), "html/trace_viewer_full.html"), encoding="utf-8") as f:
            sub["trace_viewer_full"] = f.read()
        sub["json_data"] = self.generate_json(file_info=file_info)
        sub["json_data"] = sub["json_data"].replace("</script>", "<\\/script>")

        if self.verbose > 0:
            print("Generating HTML report")

        return Template(tmpl).substitute(sub)

    def save(self, output_file="result.html", file_info=False):
        file_type = output_file.split(".")[-1]

        if self.verbose > 0:
            print("Saving report to {} ...".format(os.path.abspath(output_file)))

        if file_type == "html":
            with open(output_file, "w", encoding="utf-8") as f:
                f.write(self.generate_report(file_info=True))
        elif file_type == "json":
            data = self.generate_json(allow_binary=True, file_info=file_info)
            if type(data) is bytes:
                with open(output_file, "wb") as f:
                    f.write(data)
            else:
                with open(output_file, "w", encoding="utf-8") as f:
                    f.write(data)
        elif file_type == "gz":
            data = self.generate_json(allow_binary=True, file_info=file_info)
            if type(data) is not bytes:
                data = data.encode("utf-8")
            with gzip.open(output_file, "wb") as f:
                f.write(data)
        else:
            raise Exception("Only html, json and gz are supported")

        if self.verbose > 0:
            print("Report saved.")



# Licensed under the Apache License: http://www.apache.org/licenses/LICENSE-2.0
# For details: https://github.com/gaogaotiantian/viztracer/blob/master/NOTICE.txt

import copy
import bisect
import re


class FuncTreeNode:
    name_regex = r"(.*) \((.*?):([0-9]+)\)"

    def __init__(self, event=None):
        self.filename = None
        self.lineno = None
        self.caller_lineno = -1
        self.is_python = False
        self.funcname = None
        if event is None:
            self.event = {"name": "__ROOT__"}
            self.fullname = "__ROOT__"
            self.parent = None
            self.children = []
            self.start = - (2 ** 64)
            self.end = 2 ** 64
        else:
            self.event = copy.copy(event)
            self.parent = None
            self.children = []
            self.start = self.event["ts"]
            self.end = self.event["ts"] + self.event["dur"]
            self.fullname = self.event["name"]
            m = re.match(self.name_regex, self.fullname)
            if m:
                self.is_python = True
                self.funcname = m.group(1)
                self.filename = m.group(2)
                self.lineno = int(m.group(3))
            if "caller_lineno" in self.event:
                self.caller_lineno = self.event["caller_lineno"]

    def is_ancestor(self, other):
        return self.start < other.start and self.end > other.end

    def is_same(self, other):
        return (self.fullname == other.fullname
                and len(self.children) == len(other.children)
                and all([t[0].is_same(t[1]) for t in zip(self.children, other.children)]))

    def adopt(self, other):
        new_children = []
        if self.is_ancestor(other):
            start_array = [n.start for n in self.children]
            end_array = [n.end for n in self.children]
            start_idx = bisect.bisect(start_array, other.start)
            end_idx = bisect.bisect(end_array, other.end)
            if (start_idx == end_idx + 1):
                self.children[end_idx].adopt(other)
            elif (start_idx == end_idx):
                other.parent = self
                self.children.insert(start_idx, other)
            elif (start_idx < end_idx):
                def change_parent(node):
                    node.parent = other
                new_children = self.children[start_idx:end_idx]
                # force map to run
                list(map(change_parent, new_children))
                other.children = new_children
                other.parent = self
                self.children = self.children[:start_idx] + [other] + self.children[end_idx:]
            else:  # pragma: no cover
                raise Exception("This should not be possible")
        else:
            self.parent.adopt(other)


class FuncTree:
    def __init__(self, pid=0, tid=0):
        self.root = FuncTreeNode()
        self.curr = self.root
        self.pid = pid
        self.tid = tid

    def is_same(self, other):
        return self.root.is_same(other.root)

    def add_event(self, event):
        node = FuncTreeNode(event)

        self.curr.adopt(node)
        self.curr = node

    def first_ts(self):
        return self.root.children[0].event["ts"]

    def first_node(self):
        return self.root.children[0]

    def node_by_timestamp(self, ts):
        starts = [node.start for node in self.root.children]
        idx = bisect.bisect(starts, ts)
        if idx == 0:
            return self.root.children[0]
        else:
            return self.root.children[idx - 1]

    def normalize(self, first_ts):
        for node in self.inorder_traverse():
            node.start -= first_ts
            node.end -= first_ts

    def inorder_traverse(self):
        lst = [self.root]
        while lst:
            ret = lst.pop()
            lst.extend(ret.children[::-1])
            yield ret
        return
